package org.apache.maven.plugin.resources.remote.stub;

/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

import java.util.Properties;

import org.apache.maven.model.Model;
import org.apache.maven.model.Parent;

/**
 * Stub
 */
public class ModelStub
    extends Model
{
    private static final long serialVersionUID = 1L;

    public ModelStub()
    {

    }

    @Override
    public String getVersion()
    {
        return "0.0-TEST";
    }

    @Override
    public String getModelVersion()
    {
        return "0.0-TEST";
    }

    @Override
    public String getName()
    {
        return "Test Model";
    }

    @Override
    public String getGroupId()
    {
        return "org.apache.maven.test";
    }

    @Override
    public String getPackaging()
    {
        return "jar";
    }

    @Override
    public Parent getParent()
    {
        //return new Parent();
        return null;
    }

    @Override
    public String getArtifactId()
    {
        return "maven-test-plugin";
    }

    @Override
    public Properties getProperties()
    {
        return new Properties();
    }

}
